#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "sqlite3json.h"

struct pfs {
    putfunc *pfunc;
    void *arg;
};

static void
pstr(const char *string, struct pfs *pfs)
{
    while (*string) {
	pfs->pfunc(*string, pfs->arg);
	string++;
    }
}

static void
pstrq(const char *string, struct pfs *pfs)
{
    putfunc *pfunc = pfs->pfunc;
    void *arg = pfs->arg;
    char buf[64];

    if (!string) {
	pstr("null", pfs);
	return;
    }
    pfunc('"', arg);
    while (*string) {
	switch (*string) {
	case '"':
	case '\\':
	    pfunc('\\', arg);
	    pfunc(*string, arg);
	    break;
	case '\b':
	    pfunc('\\', arg);
	    pfunc('b', arg);
	    break;
	case '\f':
	    pfunc('\\', arg);
	    pfunc('f', arg);
	    break;
	case '\n':
	    pfunc('\\', arg);
	    pfunc('n', arg);
	    break;
	case '\r':
	    pfunc('\\', arg);
	    pfunc('r', arg);
	    break;
	case '\t':
	    pfunc('\\', arg);
	    pfunc('t', arg);
	    break;
	default:
	    if ((*string < ' ' && *string > 0) || *string == 0x7f) {
		sprintf(buf, "\\u%04x", *string);
		pstr(buf, pfs);
	    } else if (*string < 0) {
		unsigned char c = string[0];
		unsigned long uc = 0;

		if (c < 0xc0) {
		    uc = c;
		} else if (c < 0xe0) {
		    if ((string[1] & 0xc0) == 0x80) {
			uc = ((c & 0x1f) << 6) | (string[1] & 0x3f);
			++string;
		    } else {
			uc = c;
		    }
		} else if (c < 0xf0) {
		    if ((string[1] & 0xc0) == 0x80 &&
			(string[2] & 0xc0) == 0x80) {
			uc = ((c & 0x0f) << 12) |
			     ((string[1] & 0x3f) << 6) | (string[2] & 0x3f);
			string += 2;
		    } else {
			uc = c;
		    }
		} else if (c < 0xf8) {
		    if ((string[1] & 0xc0) == 0x80 &&
			(string[2] & 0xc0) == 0x80 &&
			(string[3] & 0xc0) == 0x80) {
			uc = ((c & 0x03) << 18) |
			     ((string[1] & 0x3f) << 12) |
			     ((string[2] & 0x3f) << 6) |
			     (string[4] & 0x3f);
			string += 3;
		    } else {
			uc = c;
		    }
		} else if (c < 0xfc) {
		    if ((string[1] & 0xc0) == 0x80 &&
			(string[2] & 0xc0) == 0x80 &&
			(string[3] & 0xc0) == 0x80 &&
			(string[4] & 0xc0) == 0x80) {
			uc = ((c & 0x01) << 24) |
			     ((string[1] & 0x3f) << 18) |
			     ((string[2] & 0x3f) << 12) |
			     ((string[4] & 0x3f) << 6) |
			     (string[5] & 0x3f);
			string += 4;
		    } else {
			uc = c;
		    }
		} else {
		    /* ignore */
		    ++string;
		}
		if (uc < 0x10000) {
		    sprintf(buf, "\\u%04lx", uc);
		} else {
		    uc -= 0x10000;

		    sprintf(buf, "\\u%04lx", 0xdc00 | ((uc >> 10) & 0x3ff));
		    sprintf(buf, "\\u%04lx", 0xd800 | (uc & 0x3ff));
		    pstr(buf, pfs);
		}
		pstr(buf, pfs);
	    } else {
		pfunc(*string, arg);
	    }
	    break;
	}
	++string;
    }
    pfunc('"', arg);
}

int
sqlite3json(sqlite3 *db, const char *sql, putfunc *pfunc, void *arg)
{
    struct pfs pfs0, *pfs = &pfs0;
    const char *tail = sql;
    int i, nresults = 0, result = SQLITE_ERROR;

    pfs->pfunc = pfunc;
    pfs->arg = arg;
    pstr("{sql:", pfs);
    pstrq(sql, pfs);
    pstr(",results:[", pfs);
    do {
	sqlite3_stmt *stmt;
	int firstrow = 1, nrows = 0;
	char buf[256];

	++nresults;
	pstr(nresults == 1 ? "{" : ",{", pfs);
	result = sqlite3_prepare(db, tail, -1, &stmt, &tail);
	if (result != SQLITE_OK) {
doerr:
	    if (nrows == 0) {
		pstr("columns:null,rows:null,changes:0,insrowid:null", pfs);
	    }
	    pstr("error:", pfs);
	    pstrq(sqlite3_errmsg(db), pfs);
	    pfunc('}', arg);
	    break;
	}
	result = sqlite3_step(stmt);
	while (result == SQLITE_ROW || result == SQLITE_DONE) {
	    if (firstrow) {
		for (i = 0; i < sqlite3_column_count(stmt); i++) {
		    char *type;
		    pstr(i == 0 ? "columns:[" : ",", pfs);
		    pstr("{name:", pfs);
		    pstrq(sqlite3_column_name(stmt, i), pfs);
		    pstr(",decltype:", pfs);
		    pstrq(sqlite3_column_decltype(stmt, i), pfs);
		    pstr(",type:", pfs);
		    switch (sqlite3_column_type(stmt, i)) {
		    case SQLITE_INTEGER:
			type = "integer";
			break;
		    case SQLITE_FLOAT:
			type = "float";
			break;
		    case SQLITE_BLOB:
			type = "blob";
			break;
		    case SQLITE_TEXT:
			type = "text";
			break;
		    case SQLITE_NULL:
			type = "null";
			break;
		    default:
			type = "unknown";
			break;
		    }
		    pstrq(type, pfs);
		    pfunc('}', arg);
		}
		if (i) {
		    pfunc(']', arg);
		}
		firstrow = 0;
	    }
	    if (result == SQLITE_DONE) {
		break;
	    }
	    ++nrows;
	    pstr(nrows == 1 ? ",rows:[" : ",", pfs);
	    for (i = 0; i < sqlite3_column_count(stmt); i++) {
		pfunc(i == 0 ? '[' : ',', arg);
		switch (sqlite3_column_type(stmt, i)) {
		case SQLITE_INTEGER:
		case SQLITE_FLOAT:
		    pstr(sqlite3_column_text(stmt, i), pfs);
		    break;
		case SQLITE_BLOB:
		case SQLITE_TEXT:
		    pstrq(sqlite3_column_text(stmt, i), pfs);
		    break;
		case SQLITE_NULL:
		default:
		    pstr("null", pfs);
		    break;
		}
	    }
	    pstr(i == 0 ? "null]" : "]", pfs);
	    result = sqlite3_step(stmt);
	}
	if (nrows > 0) {
	    pfunc(']', arg);
	}
	result = sqlite3_finalize(stmt);
	if (result != SQLITE_OK) {
	    goto doerr;
	}
	if (nrows == 0) {
	    pstr("columns:null,rows:null", pfs);
	}
	sprintf(buf, ",changes:%d,last_insert_rowid:%lld",
		sqlite3_changes(db),
		sqlite3_last_insert_rowid(db));
	pstr(buf, pfs);
	pstr(",error:null}", pfs);
    } while (tail && *tail);
    pstr("]}", pfs);
    return result;
}

#ifdef TEST
int main(int argc, char **argv)
{
    sqlite3 *db;
    int result;
    putfunc *pfunc = (putfunc *) fputc;
    void *arg = (void *) stdout;
    struct pfs pfs0, *pfs = &pfs0;

    pfs->pfunc = pfunc;
    pfs->arg = arg;
    if (argc != 3) {
	fprintf(stderr, "usage: %s sqlite3db sql\n", argv[0]);
	exit(1);
    }
    result = sqlite3_open(argv[1], &db);
    if (result != SQLITE_OK) {
	pstr("{\nresults:", pfs);
	pstr("[{", pfs);
	pstr("error:", pfs);
	pstrq("unable to open database file", pfs);
	pstr("}]", pfs);
	exit(1);
    }
    result = sqlite3json(db, argv[2], pfunc, arg);
    sqlite3_close(db);
    return result == SQLITE_OK ? 0 : 1;
}
#endif
